@@ -27,7 +27,7 @@
 
 #include <string.h>
 #include <unistd.h>
-#include <e-util/e-time-utils.h>
+#include <libedataserver/e-time-utils.h>
 
 #include <libgnomevfs/gnome-vfs-mime-utils.h>
 
@@ -69,7 +69,7 @@ struct ECalBackendExchangePrivate {
 #define PARENT_TYPE E_TYPE_CAL_BACKEND_SYNC
 static GObjectClass *parent_class = NULL;
 
-#define d(x)
+#define d(x) 
 
 static ECalBackendSyncStatus
 is_read_only (ECalBackendSync *backend, EDataCal *cal, gboolean *read_only)
@@ -132,7 +132,7 @@ get_static_capabilities (ECalBackendSync *backend, EDataCal *cal, char **capabil
 	return GNOME_Evolution_Calendar_Success;
 }
 
-static void
+static ECalBackendSyncStatus
 load_cache (ECalBackendExchange *cbex, E2kUri *e2kuri)
 {
 	icalcomponent *vcalcomp, *comp, *tmp_comp;
@@ -144,11 +144,15 @@ load_cache (ECalBackendExchange *cbex, E2kUri *e2kuri)
 	int i;
 	struct stat buf;
 
+	uristr = e_cal_backend_get_uri (E_CAL_BACKEND (cbex));
 	cbex->priv->object_cache_file =
 		e_folder_exchange_get_storage_file (cbex->folder, "cache.ics");
+	if (!cbex->priv->object_cache_file) {
+		printf ("could not load cache for %s\n", uristr);
+		return GNOME_Evolution_Calendar_OfflineUnavailable;
+	}
 
 	/* Fixme : Try avoiding to do this everytime we come here */
-	uristr = e_cal_backend_get_uri (E_CAL_BACKEND (cbex));
 	mangled_uri = g_strdup (uristr);
 	for (i = 0; i < strlen (mangled_uri); i++) {
 		switch (mangled_uri[i]) {
@@ -166,13 +170,18 @@ load_cache (ECalBackendExchange *cbex, E2kUri *e2kuri)
 	g_free (storage_dir);
 	g_free (mangled_uri);
 
+       /* Check if the cache file is present. If it is not present the account might
+          be newly created one. It will be created while save the cache */
+       if (!g_file_test (cbex->priv->object_cache_file, G_FILE_TEST_EXISTS))
+               return GNOME_Evolution_Calendar_Success;
+
 	vcalcomp = e_cal_util_parse_ics_file (cbex->priv->object_cache_file);
 	if (!vcalcomp)
-		return;
+		return GNOME_Evolution_Calendar_InvalidObject;
 
 	if (icalcomponent_isa (vcalcomp) != ICAL_VCALENDAR_COMPONENT) {
 		icalcomponent_free (vcalcomp);
-		return;
+		return GNOME_Evolution_Calendar_InvalidObject;
 	}
 
 	kind = e_cal_backend_get_kind (E_CAL_BACKEND (cbex));
@@ -205,6 +214,7 @@ load_cache (ECalBackendExchange *cbex, E2kUri *e2kuri)
 	}
 
 	icalcomponent_free (vcalcomp);
+	return GNOME_Evolution_Calendar_Success;
 }
 
 static void
@@ -287,10 +297,10 @@ open_calendar (ECalBackendSync *backend, EDataCal *cal, gboolean only_if_exists,
 	ECalBackendExchange *cbex = E_CAL_BACKEND_EXCHANGE (backend);
 	const char *uristr;
 	ExchangeHierarchy *hier;
-	ExchangeAccount *act;
 	ExchangeAccountResult acresult;
 	const char *prop = PR_ACCESS;
 	E2kHTTPStatus status;
+	ECalBackendSyncStatus load_status;
 	E2kResult *results;
 	E2kUri *euri = NULL;
 	int nresults;
@@ -318,19 +328,23 @@ open_calendar (ECalBackendSync *backend, EDataCal *cal, gboolean only_if_exists,
 		}
 
 		euri = e2k_uri_new (uristr);
-		load_cache (cbex, euri);
+		load_status = load_cache (cbex, euri);
 		e2k_uri_free (euri);
 		g_mutex_unlock (cbex->priv->open_lock);
-		return GNOME_Evolution_Calendar_Success;
+		return load_status;
 	}
 		
 	/* Make sure we have an open connection */
 	/* This steals the ExchangeAccount from ExchangeComponent */
-	act = exchange_component_get_account_for_uri (global_exchange_component, NULL);
-	if (!exchange_account_get_context (act))
-		exchange_account_connect (act, password, &acresult);
-
 	cbex->account = exchange_component_get_account_for_uri (global_exchange_component, uristr);
+	if (!cbex->account) {
+		cbex->account = exchange_component_get_account_for_uri (global_exchange_component, NULL);
+		if (!exchange_account_connect (cbex->account, password, &acresult)) {
+			g_mutex_unlock (cbex->priv->open_lock);
+			return GNOME_Evolution_Calendar_AuthenticationFailed;
+		}
+	}
+
 	if (!cbex->account) {
 		g_mutex_unlock (cbex->priv->open_lock);
 		return GNOME_Evolution_Calendar_PermissionDenied;
@@ -338,8 +352,9 @@ open_calendar (ECalBackendSync *backend, EDataCal *cal, gboolean only_if_exists,
 
 	if (!exchange_account_get_context (cbex->account)) {
 		g_mutex_unlock (cbex->priv->open_lock);
-		return GNOME_Evolution_Calendar_RepositoryOffline; 
-	}	
+		return GNOME_Evolution_Calendar_AuthenticationFailed;
+	}
+
 
 	cbex->folder = exchange_account_get_folder (cbex->account, uristr);
 	if (!cbex->folder) {
@@ -375,11 +390,11 @@ open_calendar (ECalBackendSync *backend, EDataCal *cal, gboolean only_if_exists,
 
 	cbex->priv->read_only = ((access & MAPI_ACCESS_CREATE_CONTENTS) == 0);
 
-	load_cache (cbex, euri);
+	load_status = load_cache (cbex, euri);
 
 	g_mutex_unlock (cbex->priv->open_lock);
-
-	return GNOME_Evolution_Calendar_Success;
+	
+	return load_status;
 }
 
 static ECalBackendSyncStatus
@@ -502,6 +517,8 @@ e_cal_backend_exchange_add_object (ECalBackendExchange *cbex,
 
 	is_instance = (icalcomponent_get_first_property (comp, ICAL_RECURRENCEID_PROPERTY) != NULL);
 
+	d(printf("ecbe_add_object: is_instance: %d\n", is_instance));	
+
 	if (ecomp && ecomp->icomp && !is_instance)
 		return FALSE;
 
@@ -537,6 +554,8 @@ e_cal_backend_exchange_add_object (ECalBackendExchange *cbex,
 
 			ecomp->instances = g_list_prepend (ecomp->instances,
 						   icalcomponent_new_clone (comp));
+			if (ecomp->icomp)
+				e_cal_util_remove_instances (ecomp->icomp, rid, CALOBJ_MOD_THIS);
 		}
 	} else
 		ecomp->icomp = icalcomponent_new_clone (comp);
@@ -557,9 +576,14 @@ discard_detached_instance (ECalBackendExchangeComponent *ecomp,
 		if (icaltime_compare (inst_rid, rid) == 0) {
 			ecomp->instances = g_list_remove (ecomp->instances, inst->data);
 			icalcomponent_free (inst->data);
-			return;
+			break;
 		}
 	}
+
+	if (ecomp->icomp) /* This check should ideally not be needed. */
+		e_cal_util_remove_instances (ecomp->icomp, rid, CALOBJ_MOD_THIS);
+
+	return;
 }
 
 /**
@@ -666,35 +690,96 @@ get_exchange_comp (ECalBackendExchange *cbex, const char *uid)
 	return NULL;			
 }
 
+static void
+add_instances_to_vcal (gpointer value, gpointer user_data)
+{
+	icalcomponent *recurrence = value;
+	icalcomponent *vcalendar = user_data;
+
+	icalcomponent_add_component (
+		vcalendar,
+		icalcomponent_new_clone (recurrence));		     
+
+}
+
 static ECalBackendSyncStatus
 get_object (ECalBackendSync *backend, EDataCal *cal,
 	    const char *uid, const char *rid, char **object)
 {
 	ECalBackendExchange *cbex = E_CAL_BACKEND_EXCHANGE (backend);
 	ECalBackendExchangeComponent *ecomp;
-	ECalComponent *comp;
 	
 	d(printf("ecbe_get_object(%p, %p, uid=%s, rid=%s)\n", backend, cal, uid, rid));
 	
 	g_return_val_if_fail (uid != NULL, GNOME_Evolution_Calendar_InvalidObject);
 	/*any other asserts?*/
 	
+	*object = NULL;
 	ecomp = g_hash_table_lookup (cbex->priv->objects, uid);
-	if ((!ecomp) || (!ecomp->icomp))
+	if (!ecomp)
 		return GNOME_Evolution_Calendar_ObjectNotFound;
-	
-	/*anything on recur id here????*/
-	
-	comp = e_cal_component_new ();
-	if (e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (ecomp->icomp))) {
-		*object = e_cal_component_get_as_string (comp);
-	} else {
-		*object = NULL;
-		g_object_unref (comp);
+
+	if (!rid && (!(ecomp->icomp)))	
 		return GNOME_Evolution_Calendar_ObjectNotFound;
+
+	if (rid && *rid) {
+		GList *l;
+		struct icaltimetype inst_rid, key_rid;
+		gboolean inst_found = FALSE;
+
+		for (l = ecomp->instances; l ; l = l->next) {
+			key_rid = icaltime_from_string (rid);
+			inst_rid = icalcomponent_get_recurrenceid (l->data);
+			if (icaltime_compare (inst_rid, key_rid) == 0) {
+				inst_found = TRUE;
+				break;
+			}
+		}
+		
+		if (inst_found) {
+			*object = g_strdup (icalcomponent_as_ical_string (l->data));
+		} else {
+			/* Instance is not found. Send the master object instead */	
+			if (ecomp->icomp) {
+				icalcomponent *new_inst;
+				struct icaltimetype itt;
+
+				itt = icaltime_from_string (rid);
+				new_inst = e_cal_util_construct_instance (
+					icalcomponent_new_clone (ecomp->icomp), itt);
+				if (!new_inst)
+					return GNOME_Evolution_Calendar_ObjectNotFound;
+				
+				*object = g_strdup (icalcomponent_as_ical_string (new_inst));
+				icalcomponent_free (new_inst);
+			} else {
+				/* Oops. No instance and no master as well !! */
+				return GNOME_Evolution_Calendar_ObjectNotFound;
+			} /* Close check for master object */
+		} /* Close check for instance */
+	}/* Close check if rid is being asked */
+	else {
+		/* rid is not asked */
+		if (g_list_length (ecomp->instances) > 0) {
+			/* Ok, we have detached instances, return a VCALENDAR */
+			/* Send the whole object including all its detatched instances */
+			icalcomponent *vcalcomp;
+
+			vcalcomp = e_cal_util_new_top_level ();
+		
+			if (ecomp->icomp)
+				icalcomponent_add_component (vcalcomp,
+					icalcomponent_new_clone (ecomp->icomp));
+			g_list_foreach (ecomp->instances, (GFunc) add_instances_to_vcal, vcalcomp);
+			
+			*object = g_strdup (icalcomponent_as_ical_string (vcalcomp));
+			icalcomponent_free (vcalcomp);	
+		} else {
+			/* There are no detached instances. Send only the master object */
+			*object = g_strdup (icalcomponent_as_ical_string (ecomp->icomp));
+		}
 	}
-	
-	g_object_unref (comp);
+
 	return GNOME_Evolution_Calendar_Success;
 }
 
@@ -799,6 +884,30 @@ typedef struct {
 	ECalBackend *backend;
 	icaltimezone *default_zone;
 } MatchObjectData;
+static void
+match_recurrence_sexp (gpointer data, gpointer user_data)
+{
+	icalcomponent *icomp = data;
+	MatchObjectData *match_data = user_data;
+	ECalComponent *comp = NULL;
+	char* comp_as_string = NULL;
+	
+	d(printf("ecbe_match_recurrence_sexp(%p, %p)\n", icomp, match_data));
+
+	if (!icomp || !match_data)
+		return;
+
+	comp = e_cal_component_new ();
+	e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (icomp));
+
+	if ((!match_data->search_needed) ||
+	    (e_cal_backend_sexp_match_comp (match_data->obj_sexp, comp, match_data->backend))) {
+		comp_as_string = e_cal_component_get_as_string (comp);
+		match_data->obj_list = g_list_append (match_data->obj_list, comp_as_string);
+		d(printf ("ecbe_match_recurrence_sexp: match found, adding \n%s\n", comp_as_string));
+	}
+	g_object_unref (comp);
+}
 
 static void
 match_object_sexp (gpointer key, gpointer value, gpointer data)
@@ -806,23 +915,31 @@ match_object_sexp (gpointer key, gpointer value, gpointer data)
 	ECalBackendExchangeComponent *ecomp = value;
 	MatchObjectData *match_data = data;
 	ECalComponent *comp;
-	
-	comp = e_cal_component_new ();
-	e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (ecomp->icomp));
 
-	if ((!match_data->search_needed) ||
-	    (e_cal_backend_sexp_match_comp (match_data->obj_sexp, comp, match_data->backend))) {
-		match_data->obj_list = g_list_append (match_data->obj_list,
+	/* 
+	   In case of detached instances with no master object,
+	   ecomp->icomp will be NULL and hence should be skipped
+	   from matching.
+	*/
+	if (!ecomp || !match_data)
+		return;
+
+	if (ecomp->icomp) {
+		comp = e_cal_component_new ();
+		e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (ecomp->icomp));
+		
+		if ((!match_data->search_needed) ||
+		    (e_cal_backend_sexp_match_comp (match_data->obj_sexp, comp, match_data->backend))) {
+			match_data->obj_list = g_list_append (match_data->obj_list,
 						      e_cal_component_get_as_string (comp));
-
-		#if 0
-		/* match also recurrences */
-		g_hash_table_foreach (obj_data->recurrences,
-				      (GHFunc) match_recurrence_sexp,
-				      match_data);
-		#endif
+		}
+		g_object_unref (comp);
 	}
-	g_object_unref (comp);
+
+	/* match also recurrences */
+	g_list_foreach (ecomp->instances,
+			(GFunc) match_recurrence_sexp,
+			match_data);
 }
 
 static ECalBackendSyncStatus
@@ -905,6 +1022,7 @@ e_cal_backend_exchange_add_timezone (ECalBackendExchange *cbex,
 	if (!prop) 
 		return GNOME_Evolution_Calendar_InvalidObject;
 	tzid = icalproperty_get_tzid (prop);
+	d(printf("ecbe_add_timezone: tzid = %s\n", tzid));
 	if (g_hash_table_lookup (cbex->priv->timezones, tzid))
 		return GNOME_Evolution_Calendar_ObjectIdAlreadyExists;
 
@@ -926,6 +1044,9 @@ add_timezone (ECalBackendSync *backend, EDataCal *cal,
 	GNOME_Evolution_Calendar_CallStatus status;
 	icalcomponent *vtzcomp;
 
+	if (!tzobj)
+		return GNOME_Evolution_Calendar_InvalidObject;
+
 	vtzcomp = icalcomponent_new_from_string ((char *)tzobj);
 	if (!vtzcomp)
 		return GNOME_Evolution_Calendar_InvalidObject;
@@ -951,16 +1072,16 @@ set_default_timezone (ECalBackendSync *backend, EDataCal *cal,
 		      const char *tzid)
 {
 	ECalBackendExchange *cbex = E_CAL_BACKEND_EXCHANGE (backend);
-	icaltimezone *zone;
 
 	d(printf("ecbe_set_default_timezone(%p, %p, %s)\n", backend, cal, tzid));
-
-	zone = g_hash_table_lookup (cbex->priv->timezones, tzid);
-	if (zone) {
-		cbex->priv->default_timezone = zone;
-		return GNOME_Evolution_Calendar_Success;
-	} else
-		return GNOME_Evolution_Calendar_ObjectNotFound;
+	/* 
+	   We call this function before calling e_cal_open in client and
+	   hence we set the timezone directly.  In the implementation of
+	   e_cal_open, we set this timezone to every-objects that we create.
+	*/
+	   
+	cbex->priv->default_timezone = icaltimezone_get_builtin_timezone_from_tzid (tzid);
+	return GNOME_Evolution_Calendar_Success;
 }
 
 struct search_data {
@@ -1064,6 +1185,7 @@ set_mode (ECalBackend *backend, CalMode mode)
 	ECalBackendExchange *cbex;
 	ECalBackendExchangePrivate *priv;
 	ExchangeAccount *account;
+	ExchangeAccountResult acresult;
 	const char *uristr;
 	
 	cbex = E_CAL_BACKEND_EXCHANGE (backend);
@@ -1086,35 +1208,35 @@ set_mode (ECalBackend *backend, CalMode mode)
 			/* Should we check for access rights before setting this ? */
 			d(printf ("set mode to online\n"));
 			account = exchange_component_get_account_for_uri (global_exchange_component, NULL);
+			/* check if authentication is required */
 			if (!exchange_account_get_context (account))
 				e_cal_backend_notify_auth_required(backend);
-			e_cal_backend_notify_mode (backend, 
-				GNOME_Evolution_Calendar_CalListener_MODE_SET,
-				GNOME_Evolution_Calendar_MODE_REMOTE);
 
 			uristr = e_cal_backend_get_uri (E_CAL_BACKEND (backend));
 			account = exchange_component_get_account_for_uri (global_exchange_component, uristr);
-			if (!account) {
-				g_mutex_unlock (priv->set_lock);	
-				return;
-			}
-			cbex->folder = exchange_account_get_folder (account, uristr);
+			if (account)
+				cbex->folder = exchange_account_get_folder (account, uristr);
+
+			e_cal_backend_notify_mode (backend, 
+				GNOME_Evolution_Calendar_CalListener_MODE_SET,
+				GNOME_Evolution_Calendar_MODE_REMOTE);
 			/* FIXME : Test if available for read already */
 			priv->read_only = FALSE;
 			priv->mode = CAL_MODE_REMOTE;
-			/* FIXME : Check if online and check if authentication
-				is needed */
 			break;
 
 	case CAL_MODE_LOCAL:
 			d(printf ("set mode to offline\n"));
-			uristr = e_cal_backend_get_uri (E_CAL_BACKEND (backend));
-			account = exchange_component_get_account_for_uri (global_exchange_component, uristr);
-			if (!account) {
-				g_mutex_unlock (priv->set_lock);	
-				return;
+			account = exchange_component_get_account_for_uri (global_exchange_component, NULL);
+			if (account) {
+				/* Call exchange_account_connect to populate the folder hierarchy,
+				 * so that load_cache succeeds */
+				exchange_account_set_offline (account);
+				if (!exchange_account_connect (account, NULL, &acresult)) {
+					uristr = e_cal_backend_get_uri (E_CAL_BACKEND (backend));
+					cbex->folder = exchange_account_get_folder (account, uristr);
+				}
 			}
-			cbex->folder = exchange_account_get_folder (account, uristr);
 			priv->mode = CAL_MODE_LOCAL;
 			priv->read_only = TRUE;
 			e_cal_backend_notify_mode (backend, 
@@ -1267,13 +1389,30 @@ check_change_type (gpointer key, gpointer value, gpointer data)
 	ECalBackendExchangeComponent *ecomp = value;
 	struct ChangeData *change_data = data;
 	char *calobj;
-	ECalComponent *comp;
+	ECalComponent *comp = NULL;
 	char *uid = key;
+	GList *l = NULL;
+	icalcomponent *icomp = NULL;
 	
-	if (ecomp != NULL) {
+	/* 
+	   In case of detached instances with no master object,
+	   ecomp->icomp will be NULL and hence should be skipped
+	   from matching.
+	*/
+	if (!ecomp)
+		return;
+	icomp = ecomp->icomp;
+	l = ecomp->instances;
+	for (icomp = ecomp->icomp; l; icomp = l->data, l = l->next) {
+		if (!icomp)
+			continue;
 
 		comp = e_cal_component_new ();
-		e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (ecomp->icomp));
+		/*
+		  e_cal_component_set_icalcomponent does a icalcomponent_free of 
+		  previous icalcomponent before setting the new one. 
+		 */
+		e_cal_component_set_icalcomponent (comp, icalcomponent_new_clone (icomp));	
 
 		calobj = e_cal_component_get_as_string (comp);
 		switch (e_xmlhash_compare (change_data->ehash, uid, calobj)){
@@ -1289,9 +1428,8 @@ check_change_type (gpointer key, gpointer value, gpointer data)
 		}
 		
 		g_free (calobj);
-		g_object_unref (comp);
 	}
-		
+	g_object_unref (comp);		
 }
 
 struct cbe_data {
@@ -1301,7 +1439,7 @@ struct cbe_data {
 	EXmlHash *ehash;
 };
 
-static void
+static gboolean
 e_cal_backend_exchange_compute_changes_foreach_key (const char *key, const char *value, gpointer data)
 {
 	struct cbe_data *cbedata = data;
@@ -1311,16 +1449,19 @@ e_cal_backend_exchange_compute_changes_foreach_key (const char *key, const char
 	if (!ecomp) {
 		ECalComponent *comp;
 		comp = e_cal_component_new ();
+		if (ecomp->icomp)
+			e_cal_component_set_icalcomponent (comp, 
+							   icalcomponent_new_clone (ecomp->icomp));
 		if (cbedata->kind == ICAL_VTODO_COMPONENT)
 			e_cal_component_set_new_vtype (comp, E_CAL_COMPONENT_TODO);
 		else
 			e_cal_component_set_new_vtype (comp, E_CAL_COMPONENT_EVENT);
 		e_cal_component_set_uid (comp, key);
 		cbedata->deletes = g_list_prepend (cbedata->deletes, e_cal_component_get_as_string (comp));
-		
-		e_xmlhash_remove (cbedata->ehash, key);
 		g_object_unref (comp);	
+		return TRUE;
 	}
+	return FALSE;
 }
 
 /* Attachments */
@@ -1584,7 +1725,7 @@ get_changes (ECalBackendSync *backend, EDataCal *cal,
 	cbedata.kind = e_cal_backend_get_kind (E_CAL_BACKEND (cbex));
 	cbedata.deletes = NULL;
 	cbedata.ehash = ehash;
-	e_xmlhash_foreach_key (ehash, (EXmlHashFunc)e_cal_backend_exchange_compute_changes_foreach_key, &cbedata);
+	e_xmlhash_foreach_key_remove (ehash, (EXmlHashRemoveFunc)e_cal_backend_exchange_compute_changes_foreach_key, &cbedata);
 	
 	*deletes = cbedata.deletes;
 	
